/*
  *****************************************************************************

  Copyright 1993 The Regents of the University of California.
  All Rights Reserved.

  Permission to use, copy, modify and distribute this software and its
  documentation for educational, research and non-profit purposes, without
  fee, and without a written agreement is hereby granted, provided that the 
  above copyright notice and the following three paragraphs appear in all 
  copies.

  Permission to incorporate this software into commercial products may be
  obtained from the Office of Technology Licensing, 2150 Shattuck Avenue,
  Suite 510, Berkeley, CA 94720-1620.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
  EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGES.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS
  TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

  The author may be contacted via:

  US Mail:  Brian Mirtich
            387 Soda Hall
	    Computer Science Division
	    Department of Electrical Engineering and Computer Science
	    University of California
	    Berkeley, CA 94720

 Phone:     (510) 642-8149

 EMail:     mirtich@cs.berkeley.edu

 *****************************************************************************
*/



#ifndef DIST_HEADER
#define DIST_HEADER

#include "matvect.h"

/* 
  =============================================================================

  constants

  =============================================================================
*/


#define INFINITY 1.0e+20

#define EPSILON 0.0001

#define CONE_EPS 1.0e-7


/*
  flaring angles

  These are the angles by which to flare the various planes defining
  Voronoi cones.  Planes of vertex Voronoi regions are flared in,
  planes of face Voronoi regions are flared out, and planes of edge
  Voronoi regions are flared in if they correspond to a vertex on the
  edge's coboundary, or in if they correspond to a face on the coboundary.
  Stated simply, the highest dimensional feature is given priority.

  All these constant are in radians and should be nonnegative.  Also, the
  following inequalities must hold:

  FACE_FLARE > EDGE_F_FLARE
  FACE_FLARE > VERT_FLARE
  EDGE_V_FLARE > VERT_FLARE
*/

#define VERT_FLARE    0.5e-5
#define EDGE_V_FLARE  1.0e-5
#define EDGE_F_FLARE  1.0e-5
#define FACE_FLARE    2.0e-5



/*
  =============================================================================

  data structures

  1. VERTEX, EDGE, and FACE

  These are the structures for vertices, edges, and faces.  The fields which
  appear in these structures are listed below.

    tag:  This is an integer field which appears first in all three structures.
    It takes a value V, E, or F, and is provided so we can identify the
    type of an arbitrary feature (e.g. one pointed to by a pointer to void)

    name:  This string field is present in the VERTEX and FACE structures.
    Edge names are derived from the names of the endpoint vertices.

    coords & xcoords:  coords are the coordinates of a VERTEX.  When we
    transform the VERTEX to a different frame, xcoords holds the transformed
    coordinates.

    u & xu:  u is a unit vector pointing from endpoint v1 to endpoint v2 of 
    an EDGE.  xu is the transformed version when we transform the edge to
    a different frame.

    len:  The length of an EDGE.

    plane:  This field is present in the FACE structure.  It defines the
    four coefficients of the plane of the face, i.e. the plane of the
    face is given by plane[0]*x + plane[1]*y + plane[2]*z + plane[3] = 0.
    Furthermore, the first 3 components are a unit vector, which form
    the outward normal of the face.

    verts:  This is a list of vertices (in CCW order) around a FACE.

    edges:  This is a list of edges incident to a VERTEX, or the boundary
    edges (in CCW order) of a FACE.

    v1 & v2:  The startpoint end endpoint vertices of an EDGE.

    fl & fr:  The left and right faces of an EDGE, respectively.  The 
    "left" face is the one on the left hand side of an observer standing
    on v1, looking toward v2, and standing on the outside of the polyhedron.

    cone:  This field is present in all three structures, and points to
    the list of planes defining the features voronoi region.  For a 
    FACE, only the side planes are present in this list.


  2.  POLYHEDRON

  This is the basic data structure for a polyhedron.  It contains the 
  following fields.

    name:  A string; the name of the polyhedron, if any.

    verts:  The list of vertices of the polyhedron.

    edges:  The list of edges of the polyhedron.

    faces:  What a surprise -- the list of faces of the polyhedron.

    pose:  The pose field is a transformation matrix.  The coordinates of 
    a polyhedron's vertices, and hence the location of its edges and faces 
    are all defined relative to a body frame.  All of the planes defining 
    the voronoi region are also relative to this frame.  However, a 
    polyhedron is allowed to move through space, and the pose matrix 
    expresses the pose of the body frame relative to some fixed global 
    frame.  Thus, the pose matrix Px of polyhedron x transforms body 
    coordinates to "world" coordinates.  Now let x and y be two polyhedra, 
    with pose matrices Px and Py.  Suppose we want to compute Txy, a 
    transformation matrix which transforms coordinates from x's body frame 
    to y's body frame, and Tyx, it's inverse.  It's easy to show that 
    Txy = Inv(Py) * Px, and Tyx = Inv(Px) * Py.


  3. FNODE

  FNODE stands for feature node.  FNODEs are used for building lists of
  features, such as the lists of features in the POLYHEDRON structure, and
  also the list of edges in the VERTEX structure, and the lists of vertices
  and edges in the FACE structure.  An FNODE has two fields.

    f:  f points to an actual feature.  f is a union, so it may be treated 
    as a pointer to a VERTEX, EDGE, or FACE, depending on what type of
    feature list we're dealing with.  There is also an "any" field in the
    union, for when we don't care what type of feature f points to.

    next:  This points to the next FNODE in the list, or is NULL if this is
    the last one.


  4. PNODE

  PNODE stands for plane node.  PNODEs are used for building lists of
  planes which define the voronoi regions for the features of a polyhedron.
  These lists are pointed to by the cone fields of the VERTEX, EDGE, and 
  FACE structure.  PNODEs contain the following three fields.

    plane:  These are the four coefficients of the actual plane, i.e.
    the plane is given by plane[0]*x + plane[1]*y + plane[2]*z + plane[3] = 0.
    Furthermore, the first three components of plane form a unit vector
    pointing toward the "good" side of the plane.  If a point lies within
    a voronoi region, it lies on the good side of all planes defining that
    region.  If it lies on the bad sign of any of the planes, it is not in
    the voronoi region.

    nbr:  This is points to the neighboring feature which gives rise to
    the particular plane specified by the plane field.  When a point
    lies on the "bad" side of a particular plane, the nbr field points
    to the feature we should walk to when searching for the closest
    feature pair.

    next:  This points to the next PNODE in the list, or is NULL if this is
    the last one.


  =============================================================================
*/


struct planeNode {
  REAL plane[4];  
    /* plane[0] * x + plane[1] * y + plane[2] * z + plane[3] >= 0 */
  void *nbr; /* point to feature to move to if this test fails */
  struct planeNode *next;  /* if there are more planes in this cone */
};
typedef struct planeNode PNODE;


struct vertex {
  int tag;
  char name[20];
  VECT3 coords;
  VECT3 xcoords;
  struct featureNode *edges;
  PNODE *cone;
};
typedef struct vertex VERTEX;


struct face {
  int tag;
  char name[20];
  struct featureNode *verts;
  struct featureNode *edges;
  VECT4 plane;
  PNODE *cone;
};
typedef struct face FACE;  


struct edge {
  int tag;
  VERTEX *v1, *v2;
  FACE *fl, *fr;
  VECT3 u; /* unit vector from v1 -> v2 */
  VECT3 xu; /* xformed unit vector from v1 -> v2 */
  REAL len; /* length of edge */
  PNODE *cone;
};
typedef struct edge EDGE;


struct featureNode {
  union{
    VERTEX *v;
    EDGE *e;
    FACE *f;
    void *any;
  } f;
  struct featureNode *next;
};
typedef struct featureNode FNODE;


struct polyhedron {
  char name[20];
  MAT4 pose;
  FNODE *verts;
  FNODE *edges;
  FNODE *faces;
};
typedef struct polyhedron POLYHEDRON;
    

/* 
  feature types
 
  These are the values stored in the tag fields of the VERTEX, EDGE, and 
  FACE structures.  The featTag macro returns the type of an arbitrary feature.
*/
  
#define V 1
#define E 2
#define F 3
#define featTag(featurePtr) (*((int *) (featurePtr)))


/*
  memory allocation macros
*/

#define allocVertex (VERTEX *) malloc(sizeof(VERTEX))
#define allocEdge (EDGE *) malloc(sizeof(EDGE))
#define allocFace (FACE *) malloc(sizeof(FACE))
#define allocPolyhedron (POLYHEDRON *) malloc(sizeof(POLYHEDRON))
#define allocPnode (PNODE *) malloc(sizeof(PNODE))
#define allocFnode (FNODE *) malloc(sizeof(FNODE))


/*
  transformation macros

  These transform a vertex or an edge from one frame to another using the
  transformation matrix T.  Transforming an vertex involves transforming
  a single set of coordinates.  Transforming an edge involves transforming
  the coordinates of both its endpoints, as well as its direction vector
*/

#define xformVert(T, v) xformPoint(T, ((VERTEX *)(v))->coords, ((VERTEX *)(v))->xcoords)
#define xformEdge(T, e) {xformPoint(T, ((EDGE *)(e))->v1->coords, ((EDGE *)(e))->v1->xcoords); xformPoint(T, ((EDGE *)(e))->v2->coords, ((EDGE *)(e))->v2->xcoords); xformVect(T, ((EDGE *)(e))->u, ((EDGE *)(e))->xu);}


/*
  set the pose matrix of polyhedron p
*/

#define setPose(p, T) mat4copy(T, (p)->pose)



/* 
  =============================================================================

  miscellaneous macros

  =============================================================================
*/


#define SWAP(x,y,type) {type tmp; tmp = x; x = y; y = tmp;}

#define SQR(x) ((x) * (x))

#define flush fflush(stdout)



/* 
  =============================================================================

  function declarations

  =============================================================================
*/


int getWord();
int random2();
void addFeature();
void reverseFlist();
void featureName();
VERTEX *findVertex();
EDGE *findEdge();
FACE *findFace();
int numFeats();
void *nthFeat();
void *randFeat();
REAL polyRad();
EDGE *newEdge();
int loadPolyhedronLibrary();
POLYHEDRON *createPolyhedron();
void dumpPolyhedron();
void addPlane();
void flipPlane();
void tweakPlaneNorm();
void computeVertCone();
void computeFaceCone();
void buildCones();
void dumpCones();
int vertConeChk();
int edgeConeChk();
int faceConeChk();
REAL Dvv();
REAL Dev();
REAL Dfv();
REAL Dve();
REAL Dee();
REAL Dfe();
REAL Dff();
void *closestToVert();
void *closestToEdge();
void *closestToFace();
void edgeCPs();
void *closestEdgeOrVertOnFace();
void closestEdges();
void *closestToFacePlane();
int polygonCut();
int faceOverlap();
REAL vertex_vertex();
REAL vertex_edge();
REAL edge_edge();
REAL vertex_face();
REAL edge_face();
REAL face_face();
REAL closestFeatures();
REAL closestFeaturesInit();

#endif
